"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Server = void 0;
const net_1 = require("net");
const stream_1 = require("stream");
const dgram_1 = require("dgram");
const dns_1 = __importDefault(require("dns"));
const util_1 = require("util");
const debug = require('debug')('lgou2w:minecraft-ping');
const lookupDns = util_1.promisify(dns_1.default.lookup);
const resolveDnsSrv = util_1.promisify(dns_1.default.resolveSrv);
async function pingJava(opts) {
    if (opts.protocol && typeof opts.protocol !== 'number') {
        return Promise.reject(new Error('Invalid client protocol: ' + opts.protocol + '. (Current: ' + typeof opts.protocol + ', Expected: number)'));
    }
    let ip = opts.host;
    let port = opts.port || 25565;
    let srv;
    if (!net_1.isIPv4(ip) && !opts.noSrv) {
        const target = '_minecraft._tcp.' + ip;
        debug(':resolve dns srv from host:', target);
        try {
            const addresses = await resolveDnsSrv(target);
            if (addresses.length > 0) {
                srv = addresses[0];
                ip = srv.name;
                port = srv.port;
                debug(':resolved srv:', srv);
            }
        }
        catch (err) {
            debug(':resolve failed:', err);
        }
    }
    return new Promise((resolve, reject) => {
        debug(':ready to connect');
        try {
            let end = 0;
            const start = Date.now();
            const socket = new net_1.Socket();
            const reader = new ChunkReader();
            socket.setNoDelay(true);
            socket.pipe(reader);
            socket.once('error', reject);
            reader.on('data', (data) => {
                debug(':read response');
                socket.destroy();
                let [packetId, offset] = readVarInt(data, 0);
                [, offset] = readVarInt(data, offset);
                if (packetId !== 0) {
                    reject(new Error('Bad packet id: ' + packetId));
                    return;
                }
                const response = data.slice(offset, data.length).toString('utf8');
                const status = JSON.parse(response);
                status.host = opts.host;
                status.port = socket.remotePort || port;
                status.ipv4 = socket.remoteAddress || opts.host;
                status.latency = end - start;
                status.srv = srv;
                resolve(status);
            });
            opts.timeout !== undefined && debug(':timeout:', opts.timeout);
            opts.timeout !== undefined && socket.setTimeout(opts.timeout, () => {
                socket.destroy();
                reject(new Error('Socket timeout'));
            });
            debug(':connecting...');
            socket.connect(port, ip, () => {
                debug(':connected, sending packet...');
                end = Date.now();
                const hostBuf = Buffer.from(ip);
                const portBuf = Buffer.alloc(2);
                portBuf.writeUInt16BE(port);
                const handshakeBuf = Buffer.concat([
                    writeVarInt(0),
                    writeVarInt(opts.protocol || 754),
                    writeVarInt(hostBuf.length),
                    hostBuf,
                    portBuf,
                    writeVarInt(1)
                ]);
                socket.write(Buffer.concat([writeVarInt(handshakeBuf.length), handshakeBuf]));
                socket.write(Buffer.from([1, 0]));
            });
        }
        catch (err) {
            reject(err);
        }
    });
}
function writeVarInt(value) {
    const bytes = [];
    do {
        let b = value & 127;
        value >>>= 7;
        if (value !== 0)
            b |= 128;
        bytes.push(b);
    } while (value !== 0);
    return Buffer.from(bytes);
}
function readVarInt(buf, offset) {
    let numRead = 0;
    let result = 0;
    let b = 0;
    do {
        b = buf.readUInt8(offset++);
        result |= (b & 127) << (7 * numRead++);
        if (numRead > 5)
            throw new Error('VarInt too big');
    } while ((b & 128) !== 0);
    return [result, offset];
}
class ChunkReader extends stream_1.Transform {
    constructor() {
        super(...arguments);
        this.buffer = Buffer.alloc(0);
        this.transforming = false;
    }
    async _transform(chunk, encoding, callback) {
        this.buffer = Buffer.concat([this.buffer, chunk]);
        this.transforming = true;
        let offset = 0;
        let length;
        while (true) {
            const packetStart = offset;
            try {
                [length, offset] = readVarInt(this.buffer, offset);
            }
            catch (err) {
                break;
            }
            if (offset + length > this.buffer.length) {
                offset = packetStart;
                break;
            }
            try {
                this.push(this.buffer.slice(offset, offset + length));
            }
            catch (err) {
                return this.destroy(err);
            }
            offset += length;
            await Promise.resolve();
        }
        this.buffer = this.buffer.slice(offset);
        this.flushCallback && this.flushCallback();
        this.transforming = false;
        callback();
    }
    flush(callback) {
        if (this.transforming)
            this.flushCallback = callback;
        else
            callback();
    }
}
async function pingLegacyB18To13(opts) {
    return pingLegacy(Buffer.from([0xFE]), opts);
}
async function pingLegacy14To15(opts) {
    return pingLegacy(Buffer.from([0xFE, 0x01]), opts);
}
async function pingLegacy16(opts) {
    const hostBuf = writeUTF16BE(opts.host);
    const portBuf = Buffer.alloc(4);
    const payloadLenBuf = Buffer.alloc(2);
    payloadLenBuf.writeUInt16BE(hostBuf.length + 7, 0);
    portBuf.writeInt32BE(opts.port || 25565, 0);
    return pingLegacy(Buffer.concat([
        Buffer.from([
            0xFE,
            0x01,
            0xFA,
            0x00, 0x0B,
            0x00, 0x4D, 0x00, 0x43, 0x00, 0x7C, 0x00, 0x50, 0x00, 0x69, 0x00,
            0x6E, 0x00, 0x67, 0x00, 0x48, 0x00, 0x6F, 0x00, 0x73, 0x00, 0x74
        ]),
        payloadLenBuf,
        Buffer.from([0x4E]),
        hostBuf,
        portBuf
    ]), opts);
}
async function pingLegacy(packet, opts) {
    const ip = opts.host;
    const port = opts.port || 25565;
    return new Promise((resolve, reject) => {
        debug(':ready to connect');
        try {
            let end = 0;
            const start = Date.now();
            const socket = new net_1.Socket();
            socket.setNoDelay(true);
            socket.once('error', reject);
            socket.on('data', (data) => {
                debug(':read response');
                socket.destroy();
                const response = readLegacyResponse(data, reject);
                if (!response)
                    return;
                const status = {};
                status.host = opts.host;
                status.port = socket.remotePort || port;
                status.ipv4 = socket.remoteAddress || opts.host;
                status.latency = end - start;
                status.version = response.version;
                status.players = response.players;
                status.description = response.description;
                resolve(status);
            });
            opts.timeout !== undefined && debug(':timeout:', opts.timeout);
            opts.timeout !== undefined && socket.setTimeout(opts.timeout, () => {
                socket.destroy();
                reject(new Error('Socket timeout'));
            });
            debug(':connecting...');
            socket.connect(port, ip, () => {
                debug(':connected, sending packet...');
                end = Date.now();
                socket.write(packet);
            });
        }
        catch (err) {
            reject(err);
        }
    });
}
function readLegacyResponse(buf, reject) {
    const packetId = buf.readUInt8(0);
    if (packetId !== 0xFF) {
        return reject(new Error(`Bad packet id: ${packetId}`));
    }
    const payloadLen = buf.readUInt16BE(1);
    const payload = readUTF16BE(buf.slice(3));
    if (payload.length !== payloadLen) {
        return reject(new Error(`Bad packet payload length. (expected: ${payloadLen}, current: ${payload.length})`));
    }
    if (payload[0] === '\xa7' && payload[1] === '1') {
        const pairs = payload.slice(3).split('\0');
        if (pairs.length !== 5) {
            return reject(new Error(`Bad packet payload entries: ${pairs} (expected: 5, current: ${pairs.length}`));
        }
        return {
            version: { protocol: parseInt(pairs[0]), name: pairs[1] },
            players: { max: parseInt(pairs[4]), online: parseInt(pairs[3]) },
            description: pairs[2]
        };
    }
    else {
        const pairs = payload.split('\xa7');
        if (pairs.length !== 3) {
            return reject(new Error(`Bad packet payload entries: ${pairs} (expected: 3, current: ${pairs.length}`));
        }
        return {
            version: { protocol: 0, name: '' },
            players: { max: parseInt(pairs[2]), online: parseInt(pairs[1]) },
            description: pairs[0]
        };
    }
}
function writeUTF16BE(str) {
    const buf = Buffer.from(str, 'ucs2');
    for (let i = 0; i < buf.length; i += 2) {
        const tmp = buf[i];
        buf[i] = buf[i + 1];
        buf[i + 1] = tmp;
    }
    return buf;
}
function readUTF16BE(buf) {
    if (buf.length <= 0)
        return '';
    const data = Buffer.alloc(buf.length + 1);
    let j = 0;
    for (let i = 0; i < buf.length; i += 2, j += 2) {
        data[j] = buf[i + 1];
        data[j + 1] = buf[i];
    }
    return data.slice(0, j).toString('ucs2');
}
async function pingBedrock(opts) {
    let ip = opts.host;
    const port = opts.port || 19132;
    if (!net_1.isIPv4(ip)) {
        debug(':resolve dns from host:', ip);
        const address = await lookupDns(opts.host);
        ip = address.address;
        debug(':resolved: %s -> %s', opts.host, ip);
    }
    return new Promise((resolve, reject) => {
        debug(':ready to listen');
        try {
            let timeout;
            const start = Date.now();
            const client = dgram_1.createSocket('udp4');
            client.once('error', reject);
            client.on('message', (msg, rinfo) => {
                debug(':read response');
                client.close();
                timeout && clearTimeout(timeout);
                const pong = readBedrockPacketPong(msg, reject);
                if (!pong)
                    return;
                const status = {};
                status.host = opts.host;
                status.port = rinfo.port || port;
                status.ipv4 = rinfo.address || ip;
                status.latency = Date.now() - start;
                status.description = pong.serverName;
                status.version = {
                    name: pong.serverVersion,
                    protocol: parseInt(pong.protocolVersion)
                };
                status.players = {
                    online: parseInt(pong.online),
                    max: parseInt(pong.max)
                };
                status.serverId = pong.serverId;
                status.gameId = pong.gameId;
                status.payload = pong.payload;
                resolve(status);
            });
            client.on('listening', () => {
                opts.timeout !== undefined && debug(':timeout:', opts.timeout);
                timeout = opts.timeout === undefined
                    ? undefined
                    : setTimeout(() => {
                        client.close();
                        reject(new Error('Timeout'));
                    }, opts.timeout);
                debug(':listened, sending packet...');
                const packet = writeBedrockPacketPing(start);
                client.send(packet, 0, packet.length, port, ip);
            });
            debug(':listening...');
            client.bind();
        }
        catch (err) {
            reject(err);
        }
    });
}
const BedrockEditionPacketMagic = [
    0x00, 0xFF, 0xFF, 0x00, 0xFE, 0xFE, 0xFE, 0xFE,
    0xFD, 0xFD, 0xFD, 0xFD, 0x12, 0x34, 0x56, 0x78
];
function writeBedrockPacketPing(timestamp) {
    const timestampBuf = Buffer.alloc(8);
    timestampBuf.writeBigInt64BE(BigInt(timestamp));
    return Buffer.concat([
        Buffer.from([0x01]),
        timestampBuf,
        Buffer.from(BedrockEditionPacketMagic)
    ]);
}
function readBedrockPacketPong(buf, reject) {
    const packetId = buf.readInt8(0);
    if (packetId !== 0x1C) {
        return reject(new Error(`Bad packet id: ${packetId}`));
    }
    const timestamp = buf.readBigInt64BE(1).toString();
    const serverId = buf.readBigInt64BE(9).toString() + 'n';
    if (buf.compare(Buffer.from(BedrockEditionPacketMagic), 0, BedrockEditionPacketMagic.length, 17, 33) !== 0) {
        return reject(new Error(`Bad packet magic data: ${buf.slice(17, 33).toString('hex')}`));
    }
    const payloadLen = buf.readInt16BE(33);
    const payload = buf.slice(35, payloadLen).toString().split(/;/g);
    if (payload.length < 6) {
        return reject(new Error(`Bad packet payload entries: ${payload}. (expected: 6, current: ${payload.length})`));
    }
    const [gameId, serverName, protocolVersion, serverVersion, online, max] = payload;
    return {
        timestamp,
        serverId,
        gameId,
        serverName,
        protocolVersion,
        serverVersion,
        online,
        max,
        payload
    };
}
var Server;
(function (Server) {
    Server[Server["LEGACY_B1_8_TO_1_3"] = 0] = "LEGACY_B1_8_TO_1_3";
    Server[Server["LEGACY_1_4_TO_1_5"] = 1] = "LEGACY_1_4_TO_1_5";
    Server[Server["LEGACY_1_6"] = 2] = "LEGACY_1_6";
    Server[Server["JAVA"] = 3] = "JAVA";
    Server[Server["BEDROCK"] = 10] = "BEDROCK";
})(Server || (Server = {}));
exports.Server = Server;
async function ping(server, hostOrOpts) {
    debug(':ping server:', Server[server]);
    const opts = parseOptions(hostOrOpts);
    if (!opts.host)
        throw new Error('Invalid server host: ' + opts.host);
    switch (server) {
        case Server.LEGACY_B1_8_TO_1_3: return pingLegacyB18To13(opts);
        case Server.LEGACY_1_4_TO_1_5: return pingLegacy14To15(opts);
        case Server.LEGACY_1_6: return pingLegacy16(opts);
        case Server.JAVA: return pingJava(opts);
        case Server.BEDROCK: return pingBedrock(opts);
        default: throw new Error('Invalid server type: ' + server);
    }
}
function parseOptions(hostOrOpts) {
    debug(':parse options:', hostOrOpts);
    if (!hostOrOpts)
        throw new Error('Invalid server host or options: ' + hostOrOpts);
    if (typeof hostOrOpts === 'object')
        return hostOrOpts;
    else if (typeof hostOrOpts !== 'string')
        throw new Error('Invalid server host or options type: ' + typeof hostOrOpts);
    const [host, port] = hostOrOpts.split(':', 2);
    const _port = port ? parseInt(port) : undefined;
    if (_port !== undefined && isNaN(_port))
        throw new Error('Invalid server port value: ' + port);
    debug(':parsed host: %s, port: %s', host, _port);
    return {
        host,
        port: _port
    };
}
exports.default = ping;
//# sourceMappingURL=index.js.map